// Portrait of a Digital Weapon
// Mac Pierce 2021
 
//----------------------------- Production Notes -------------------------------------------------
/*--- VERSION: --
 * - V0-01 - 210301 - Versioned off of 14seg. character code. Bodged it together.
 * - v0-02 - 210304 - Configured SD Card with text content. Compiled successfully. 
 * -- Updated segment binary sequence.
 * -- Built and verified USB enable works & programmed animation.
 * - v0-03 - 210305 - Updated SD Read to work
 * -- Incoporated White Seg Text
 * -- Added SFX ;)
 * - v0-04 - 210307 - Updates...
 * -- Updated SFX #2, changed chance on both.
 * -- Started building ring animations.
 * -- Mostly finished the animations.
 * - v0-05 - 210309 - Updates
 * -- Decreased frequency of USB glitch (995->998)
 * -- Started building write to log file to capture current LED read position
 * -- Fixed - Have to delete the file each time, but the piece now logs the current read #.
 * - v0-06 - 210310 - Tuning
 * -- Added in step to Seg Assignments to convert lowercase letter to uppercase
 * -- Color balanced the two LED types
 * - v1-00 - 210310
 * 
 * 
 * --- TODO: --
 * 
 * --- NOTES ---
 * Data order on the 16seg fixture is: 
 * - 3(16),10(16),15(7),11(15),8(7),4(14),1(7),5(16),12(16),16(7),13(15),9(7),6(15),2(7),7(16),14(16)
 * - Total of 198 LEDs
 * All LED lines have a first LED as a logic level shifter
 * USB Enable is wired into pin 12, needs to be pulled low
 * Code file on SD has to be named "text.txt"
 * if the log and text files are available, both LEDs on the back get turned on. 
 * Debug LEDs Notes -
 * -- Top LED:
 * -- Bottom LED:
 * --- Green - Log file loaded.
 * --- Red - Log file did not load.
 */ 
 
//----------------------------- Project Code -------------------------------------------------
#include <SPI.h>
#include <SD.h>
#include <Adafruit_NeoPixel.h>

#define LEDPINUSBRING 5
#define LEDPIN16SEG 10
#define USBENABLEPIN 12
#define SDPIN 4

#define segCount 16
#define segLEDCount 199 // 1 initial LED + 198 fixture LEDs
#define USBRingLEDS 17 // 1 initial LED + 16 ring LEDs

#define LEDBRT 60 // The LED Brightness - Kept low for powering off of USB.
#define S16R 255 // Color correction upper limit for red on the Segment Display
#define S16G 200 // Color correction upper limit for green on the Segment Display
#define S16B 140 // Color correction upper limit for blue on the Segment Display
#define USBR 255 // Color correction upper limit for red on the USB Ring
#define USBG 255 // Color correction upper limit for green on the USB Ring
#define USBB 102 // Color correction upper limit for blue on the USB Ring
#define USBBRIGHT 0.50 // Brightness correction for the USB Ring

Adafruit_NeoPixel SegChar16 = Adafruit_NeoPixel(segLEDCount, LEDPIN16SEG, NEO_GRB + NEO_KHZ800);
Adafruit_NeoPixel USBRing = Adafruit_NeoPixel(USBRingLEDS, LEDPINUSBRING, NEO_GRB + NEO_KHZ800);

File textFile;
File logFile;

bool activeLEDs[segLEDCount];
int activeRingLEDs[USBRingLEDS];
int segLEDs[segCount] = {16,16,7,15,7,15,7,16,16,7,15,7,15,7,16,16};
int charLEDsColor[] = {S16R,S16G,S16B};
int i = 0;
char activeChar;
bool usbEnable = false;
long sdSize;
long cRead;
int sFX;
int ringFX = 0;

//Timing stuff
int ledRefreshInterval = 125; // How fast the effects move
int ledRingRefreshInterval = 125; // How fast the effects move on the rign
int autoRefreshInterval = 125; // How fast the auto effects move
long ledPreviousMillis1 = 0;
long ledPreviousMillis2 = 0; 
long autoPreviousMillis = 0;
long inputDelay = 0; // If a function is supposed to move slower it adds onto this variable

bool debug = false; // 

//-----------------------------Initializes the sketch & checks for Serial-------------------------------------------------
void setup() {
  SegChar16.setBrightness(LEDBRT);
  USBRing.setBrightness(LEDBRT*USBBRIGHT);
  
  SegChar16.begin();
  USBRing.begin();
  
  SegChar16.show();
  USBRing.show();

  pinMode(SDPIN,OUTPUT);
  pinMode(USBENABLEPIN,INPUT);
  digitalWrite(USBENABLEPIN,HIGH); // Sets the pin high by default, so if it's not grounded it wont enable unprompted.

  charAssign('\0');
  
  SD.begin(SDPIN);
  textFile = SD.open("text.txt",FILE_READ); // Where the text is read from --------------------------------
  logFile = SD.open("log.txt",FILE_READ); // Where the current position of reading is held --------------
  
  if(textFile){
    sdSize = textFile.size();
  }
  
  if(Serial){
    Serial.begin(115200);
  }
  if(Serial){ //Initializes a serial connection if available
    Serial.println("Portrait of a Digital Weapon Connected");
    Serial.println("Mac Pierce 2021");
    Serial.println();
    if(textFile){
      Serial.println("Text loaded successfully fron SD");
    }
    else{
      
    }
  }
}

//--------------------------- Where it plays -------------------------------------------------------------------------------
void loop(){
  while(digitalRead(USBENABLEPIN)){ //Checks to see if the USB is plugged in to enable the piece
    if(usbEnable){
      if(logFile){
        logFlush(cRead);
      }
      textFile.close();
      
      SegChar16.clear();
      USBRing.clear();
      SegChar16.show();
      USBRing.show();
    }
    usbEnable = false;
  }
  if(!usbEnable){ //Checks to see if the USB has been plugged in, and played the start animation if so.
    startAnimation();
    if(!textFile){
      textFile = SD.open("text.txt",FILE_READ);
    }
    if(!logFile){
      logFile = SD.open("log.txt",FILE_READ);
    }
    if(logFile){ // If the log file is available
      USBRing.setPixelColor(0,0,50,0); // Sets the bottom debug LED to green if there's no problems reading the log file
      logFile.seek(0);
      String pRead = logFile.readStringUntil('\r');
      cRead = pRead.toInt();  
      textFile.seek(cRead); // Reads the log file to see where to start reading of the text file should start
      logFile.close();
    }
    else{
      USBRing.setPixelColor(0,50,0,0); // Sets the bottom debug LED to red if there's an error reading from the log file
    }
  }
  
  unsigned long currentMillis = millis();
  int z = 0;
  
  if(currentMillis - autoPreviousMillis > autoRefreshInterval + inputDelay){ // If 
    autoPreviousMillis = currentMillis;
    inputDelay = 125;
    
    if(ringFX >= 1){
      z = 0;
    }
    else{
      z = int(random(1000));
      sFX = 0;
    }

    if(!textFile){
      charAssign('\0');
      sFX = 0;
      }
    else if(z==500){ // ------------ ;)
      charAssign('\0');
      sFX = 1;
    }
    else if(z==750){ //-------------- :0
      charAssign('\0');
      sFX = 2;
    }
    else{
      if(z >= 997){ // 1/250 chance of happening
        ringFX = 1;
        for(int x = 1; x <= USBRingLEDS -1; x++){
          activeRingLEDs[x] = x-1-8;
        }
      }
      if(textFile.available()){
        char sign = textFile.read(); // *************************** Keep a count of active char? ******************************
        cRead = textFile.position();
        logFlush(cRead);
        
        if(cRead >= sdSize){ //Checks to see if it's at the end of the file, resets to start it if so.
          textFile.seek(0);
        }
        charAssign(sign);
      }
    }
  }
  if(currentMillis - ledPreviousMillis2 > ledRingRefreshInterval) {
    ledPreviousMillis2 = currentMillis;
    ringAnimations(ringFX);
  }
  if(currentMillis - ledPreviousMillis1 > ledRefreshInterval) {
    ledPreviousMillis1 = currentMillis;
    charAnimations(sFX);
  }
  SegChar16.show();
  USBRing.show();
}



//----------------------------- This handles setting colors of the char fixture --------------------------
void charAnimations(int z){
  if(z<=0){
      for(int x = 1; x <= segLEDCount -1; x++){ // Pushes Data to the segment fixture LEDs, except the first one.
      if(activeLEDs[x]){
        if(z < 0){ //When sFX is set to -1 via the ring animation, it triggers the sparkling red
          SegChar16.setBrightness(LEDBRT*2);
          charLEDsColor[0] = random(S16R);
          charLEDsColor[1] = random(S16G)/5;
          charLEDsColor[2] = random(S16B)/5;
          ledRefreshInterval = int(random(50))+5;
        }
        else{
          int cVal = charLEDsColor[0]+charLEDsColor[1]+charLEDsColor[2];
          if(cVal <= S16R+S16G+S16B){
            if(charLEDsColor[0] <= S16R-1){
              charLEDsColor[0]+=1;
            }
            if(charLEDsColor[1] <= S16G-1){
              charLEDsColor[1]+=1;
            }
            if(charLEDsColor[2] <= S16B-1){
              charLEDsColor[2]+=1;
            }
            ledRefreshInterval = 125;
            SegChar16.setBrightness(map(cVal,0,S16R+S16G+S16B,LEDBRT*2,LEDBRT));
          }
          
          else{
            ledRefreshInterval = 125;
          } 
        }
        SegChar16.setPixelColor(x,charLEDsColor[0],charLEDsColor[1],charLEDsColor[2]);
      }
      else{
        SegChar16.setPixelColor(x,0,0,0); //Sets the pixel to off
      }
    }
  }
  else if(z==1){ //SFX #1
    int ledPos = 1;
    for(int x = 0;x <= segCount -1; x++){ //Sees how many segments there are
      for(int y = 0; y <= segLEDs[x]-1; y++){ //Reads how many LEDs per segment
        if(x==0||x==5||x==6||x==7){
          SegChar16.setPixelColor(ledPos,0,0,255); // really blue
        }
        else if(x==2||x==4||x==9||x==11||x==13){
          SegChar16.setPixelColor(ledPos,S16R,S16G,S16B);// white
        }
        else{
          SegChar16.setPixelColor(ledPos,S16R,0,0);// red
        }
        ledPos+=1;
      }
    }
  }
  else if(z==2){ //SFX #2
    int ledPos = 1;
    for(int x = 0;x <= segCount -1; x++){ //Sees how many segments there are
      for(int y = 0; y <= segLEDs[x]-1; y++){ //Reads how many LEDs per segment
        if(x==0||x==1||x==4||x==11||x==14||x==15){
          SegChar16.setPixelColor(ledPos,S16R,S16G,S16B); // white
          
        }
        else{
          SegChar16.setPixelColor(ledPos,0,0,255); // really blue
        }
        ledPos+=1;
      }
    }
  }
}

//----------------------------- This handles setting colors of the ring --------------------------
void ringAnimations(int z){
  if(z==0){
    for(int x = 1; x <= USBRingLEDS -1; x++){ // First ring effect - pushes white to the LEDs
      if(activeRingLEDs[x]<=250){
        activeRingLEDs[x]+=5;
        ledRingRefreshInterval = 10;
      }
      else{
        ledRingRefreshInterval = 125;
      }
      int y = activeRingLEDs[x];
      USBRing.setPixelColor(x,map(y,0,255,0,USBR),map(y,0,255,0,USBG),map(y,0,255,0,USBB));
    }
  }
  else if(z==1){ // The animation for the "Loading" chase, slowly speeds up"
    int y;
    for(int x = USBRingLEDS -1; x >= 1; x--){
      y = (activeRingLEDs[x]*32)-1;
      if(y <= 0){
        USBRing.setPixelColor(x,0,0,0);
      }
      else{
        USBRing.setPixelColor(x,map(y,0,255,0,USBR),map(y,0,255,0,USBG),map(y,0,255,0,USBB));
      }
      activeRingLEDs[x] --; 
      if(activeRingLEDs[x] <= -9){
        activeRingLEDs[x] = 8;
      }
    }
    ledRingRefreshInterval --;
    if(ledRingRefreshInterval <= 0){
      ledRingRefreshInterval = 5;
      ringFX = 2;
      for(int x = 1; x <= USBRingLEDS -1; x++){
        activeRingLEDs[x] = 255;
      }
    }
  }
  else if(z==2){ // The red that fades out after the chase
    int y;
    for(int x = 1; x <= USBRingLEDS -1; x++){
      y = activeRingLEDs[x];
      USBRing.setPixelColor(x,map(y,0,255,0,USBR),0,0);
      activeRingLEDs[x]-=5;
      sFX = -1;
    }
    if(y<=0){
      ringFX = 3;
      ledRingRefreshInterval = 1000;
    }
  }
  else{ // Resets the ring animations to 0
    ringFX = 0;
    sFX = 0;
    ledRingRefreshInterval = 125;
  }
}

//-------------------------------- The animation that plays when the USB is plugged in -------------------------------------
void startAnimation(){
  delay(1500);
  for(int x = 1; x <= USBRingLEDS -1; x++){
    USBRing.setPixelColor(x,USBR,USBG,USBB);
    USBRing.show();
    delay(autoRefreshInterval);
  }
  for(int y = USBR;y > 0; y--){
    for(int x = 1; x <= USBRingLEDS -1; x++){
      USBRing.setPixelColor(x,y,0,0);
    }
    USBRing.show();
    delay(5);
  }
  usbEnable = true;
}

// -------------------------- Flushes the current data of the log file and writes the new log file. -------------------------
void logFlush(long x){ 
  SD.remove("log.txt");
  logFile = SD.open("log.txt",FILE_WRITE);
  logFile.println(x);
  logFile.close();
}

//---------------------------- Assigns the LEDs a boolean value depending on their segment ----------------------------------
void charAssign(char character){
  if(isAlphaNumeric(character)){ // Converts all lowercase letters to uppercase
    if(isLowerCase(character)){
      character -= ('a'-'A');
    }
  }
  String segments;
  switch (character){
    case '\0': //----------- Sending a Null char will set it all on ------------
      segments = "1111111111111111";
      break;
    case 'A': //---------------- The boolean sequence of each character segment --------------------
      segments="1100101000010111"; 
      break;
    case 'B':
      segments="0010001111010111";
      break;
    case 'C':
      segments="1110001001000100";
      break;
    case 'D':
      segments="0010001111000111";
      break;
    case 'E':
      segments="1110101001000100";
      break;
    case 'F':
      segments="1100101000000100";
      break;
    case 'G':
      segments="1110001001010101";
      break;
    case 'H':
      segments="1100100000010011";
      break;
    case 'I':
      segments="0010001111000100";
      break;
    case 'J':
      segments="0110000001000011";
      break;
    case 'K':
      segments="1100100000101000";
      break;
    case 'L':
      segments="1110000001000000";
      break;
    case 'M':
      segments="1100010000001011";
      break;
    case 'N':
      segments="1100010000100011";
      break;
    case 'O':
      segments="1110001001000111";
      break;
    case 'P':
      segments="1100101000010110";
      break;
    case 'Q':
      segments="1110001001100111";
      break;
    case 'R':
      segments="1100101000110110";
      break;
    case 'S':
      segments="1010101001010101";
      break;
    case 'T':
      segments="0000001110000100";
      break;
    case 'U':
      segments="1110000001000011";
      break;
    case 'V':
      segments="1101000000001000";
      break;
    case 'W':
      segments="1101000000100011";
      break;
    case 'X':
      segments="0001010000101000";
      break;
    case 'Y':
      segments="1010100001010011";
      break;
    case 'Z':
      segments="0011001001001100";
      break;
    case '0':
      segments="1111001001001111";
      break;
    case '1':
      segments="0000000000001011";
      break;
    case '2':
      segments="0110101001010110";
      break;
    case '3':
      segments="0010001001010111";
      break;
    case '4':
      segments="1000100000010011";
      break;
    case '5':
      segments="1010101001100100";
      break;
    case '6':
      segments="1110101001010101";
      break;
    case '7':
      segments="0000001000000111";
      break;
    case '8':
      segments="1110101001010111";
      break;
    case '9':
      segments="1010101001010111";
      break;
    case '/':
      segments="0001000000001000";
      break;
    case ',':
      segments="0001000000000000";
      break;
    case ':':
      segments="0000000001010000";
      break;
    case ';':
      segments="0001001000000000";
      break;
    case '=':
      segments="0000101000000000";
      break;
    case '+':
      segments="0000100110010000";
      break;
    case '{':
      segments="0011011000000000";
      break;
    case '}':
      segments="0000000001101100";
      break;
    case '(':
      segments="1110001000000000";
      break;
    case ')':
      segments="0000000001000111";
      break;
    case '[':
      segments="0010001110000000";
      break;
    case ']':
      segments="0000000111000100";
      break;
    case '.':
      segments="0000000001000000";
      break;
    case '?':
      segments="0000001111000100";
      break;
    case '!':
      segments="0000000111000000";
      break;
    case '*':
      segments="0001110110111000";
      break;
    case '@':
      segments="0110111101100000";
      break;
    case '#':
      segments="0010100111010011";
      break;
    case '$':
      segments="1010101111010101";
      break;
    case '%':
      segments="1001101111011001";
      break;
    case '-':
      segments="0000100000000000";
      break;
    case '_':
      segments="0010000001000000";
      break;
    case '|':
      segments="0000000110000000";
      break;
    case '^':
      segments="0001000000100000";
      break;
    case '>':
      segments="0001010000000000";
      break;
    case '<':
      segments="0000000000101000";
      break;
    default: //----------- Will default to an empty character if no useable chars are sent ----------
      segments="0000000000000000";
      break;
  }
  if(Serial){
    Serial.println(segments + " - Combination for: " + character);
  }
  //-----------------------Assigns the individual LED availability------------------------------
  int ledPos = 1; //Holds the position of where the loop is currently in the boolean array
  for(int x = 0;x <= segments.length()-1; x++){ //Sees how many segments there are
    for(int y = 0; y <= segLEDs[x]-1; y++){ //Reads how many LEDs per segment
      if(segments.charAt(x) == '1'){
        activeLEDs[ledPos] = true;
      }
      else{
        activeLEDs[ledPos] = false;
      }
      ledPos+=1;
    }
  }
}
